@@ -12,20 +12,23 @@ extremely handy for recovering the syntactic structure of actual rules, so that'
 ========================================
 
 """
-import re, os, collections, typing
-from ..support import foundation, compaction, interfaces
+import re, os, collections
+from typing import Optional, NamedTuple, List
+from ..support import foundation, compaction, interfaces, failureprone
 from ..parsing import automata
 from ..scanning import finite, regular, miniscan
 from . import grammar
 
 
-def compile_file(pathname, *, method, strict=False) -> dict:
-	with(open(pathname)) as fh: document = fh.read()
-	return compile_string(document, method=method).determinize().as_compact_form(filename=os.path.basename(pathname))
+class ScanRule(NamedTuple):
+	line_number: int
+	trail: Optional[int]
+	message: Optional[str]
+	parameter: Optional[str]
 
 class TextBookForm:
 	""" This provides the various views of the text-book form of scan and parse tables. """
-	def __init__(self, *, dfa: finite.DFA, scan_actions:list, parse_table: automata.DragonBookTable):
+	def __init__(self, *, dfa: finite.DFA, scan_actions:List[ScanRule], parse_table: automata.DragonBookTable):
 		self.dfa = dfa
 		self.scan_actions = scan_actions
 		self.parse_table = parse_table
@@ -42,7 +45,7 @@ class TextBookForm:
 		if dfa is None: return
 		return {
 			'dfa': compaction.compress_scanner(initial=dfa.initial, matrix=dfa.states, final=dfa.final),
-			'action': dict(zip(['message', 'parameter', 'trail', 'line_number'], zip(*self.scan_actions))),
+			'action': dict(zip(ScanRule._fields, zip(*self.scan_actions))),
 			'alphabet': {'bounds': dfa.alphabet.bounds, 'classes': dfa.alphabet.classes,}
 		}
 	def compact_parser(self):
@@ -57,7 +60,7 @@ class TextBookForm:
 			'terminals': table.terminals,
 			'nonterminals': table.nonterminals,
 			'breadcrumbs': [symbol_index[s] for s in table.breadcrumbs],
-			'rule': encode_parse_rules(table.rule_table, table.constructors, table.rule_origin),
+			'rule': encode_parse_rules(table.rule_table, table.constructors, table.rule_provenance),
 		}
 		if table.splits: form['splits'] = table.splits
 		return form
@@ -73,9 +76,9 @@ class TextBookForm:
 		if self.parse_table is not None:
 			self.parse_table.make_csv(pathstem)
 
-class IntermediateForm(typing.NamedTuple):
+class IntermediateForm(NamedTuple):
 	nfa: finite.NFA
-	scan_actions: list
+	scan_actions: List[ScanRule]
 	hfa: automata.HFA
 	parse_style:automata.ParsingStyle
 	def determinize(self) -> TextBookForm:
@@ -84,6 +87,16 @@ class IntermediateForm(typing.NamedTuple):
 	def make_dot_file(self, path): self.hfa.make_dot_file(path)
 
 
+def compile_string(document:str, *, method) -> IntermediateForm:
+	text = failureprone.SourceText(document)
+	return _compile_text(text, method=method)
+
+def compile_file(pathname, *, method, strict=False) -> dict:
+	filename = os.path.basename(pathname)
+	with(open(pathname)) as fh: text = failureprone.SourceText(fh.read(), filename=filename)
+	return _compile_text(text, method=method).determinize().as_compact_form(filename=filename)
+
+
 STRERROR = {
 	regular.BadReferenceError: "The name {0} is not defined.",
 	regular.CircularDefinitionError: "The name {0} is involved in a circular definition.",
@@ -91,26 +104,25 @@ STRERROR = {
 	regular.TrailingContextError: "Variable size for both stem and trailing context is not currently supported.",
 }
 
-def compile_string(document:str, *, method) -> IntermediateForm:
+def _compile_text(document:failureprone.SourceText, *, method) -> IntermediateForm:
 	""" This has the job of reading the specification and building the textbook-form tables. """
 	# The approach is a sort of outside-in parse. The outermost layer concerns the overall markdown document format,
 	# which is dealt with in the main body of this routine prior to determinizing and serializing everything.
 	# Each major sub-language is line-oriented and interpreted with one of the following five subroutines:
 	
-	def handle_meta_exception(e: Exception):
+	def handle_meta_exception(e: Exception, pattern_text:str):
 		if isinstance(e, regular.PatternError):
 			raise grammar.DefinitionError('At line %d: %s'%(line_number, STRERROR[type(e)].format(e.args))) from None
-		elif isinstance(e, interfaces.LanguageError):
+		else:
 			raise grammar.DefinitionError('At line %d: Malformed pattern.' % line_number) from None
-		else: raise e
 
 	def definitions():
 		name, regex = current_line_text.split(None, 1)
 		if name in env: raise grammar.DefinitionError('You cannot redefine named subexpression %r at line %d.'%(name, line_number))
 		if not re.fullmatch(r'[A-Za-z][A-Za-z_]+', name): raise grammar.DefinitionError('Subexpression %r ought to obey the rule at line %d.'%(name, line_number))
 		try: env[name] = miniscan.rex.parse(miniscan.META.scan(regex), language='Regular')
-		except Exception as e: handle_meta_exception(e)
-		assert isinstance(env[name], regular.Regular), "This would be a bug."
+		except Exception as e: handle_meta_exception(e, regex)
+		assert env[name].symbol.category in ('char_class', 'regex'), "This would be a bug: got a %r."%env[name].symbol.label
 	
 	def conditions():
 		"""
@@ -122,38 +134,40 @@ def compile_string(document:str, *, method) -> IntermediateForm:
 		name, includes = error_help.parse(current_line_text, line_number, "condition")
 		if name in condition_definitions: error_help.gripe('Re-declared scan-condition %r; this is unexpected.'%name)
 		condition_definitions[name] = includes
-	
+
+	def note_pattern(pattern_text):
+		# Now patterns that share a trail length can also share a rule ID number.
+		try: bol, expression, trail = miniscan.analyze_pattern(pattern_text, env)
+		except regular.PatternError as e:
+			handle_meta_exception(e, pattern_text)
+		else: pending_patterns[trail].append((bol, expression))
+
 	def patterns():
 		# This could be done better: a nice exercise might be to enhance the present regex parser to also
 		# grok actual scanner rules as an alternate language start-symbol; such could eliminate some of
 		# this contemptible string hackery and thereby enable things like embedded spaces where they make sense.
 		# Such would also involve hacking the metascanner bootstrap code to track paren depth and recognize
 		# the other tokens that can appear.
-		def note_pattern(pattern):
-			# Now patterns that share a trail length can also share a rule ID number.
-			try: bol, expression, trail = miniscan.analyze_pattern(pattern, env)
-			except Exception as e: handle_meta_exception(e)
-			else: pending_patterns[trail].append((bol, expression))
 		if current_line_text.endswith('|'):
-			pattern = current_line_text[:-1].strip()
-			if re.search(r'\s', pattern): raise grammar.DefinitionError('Unable to analyze pattern/same-as-next structure at line %d.')
-			note_pattern(pattern)
-			return
-		m = re.fullmatch(r'(.*?)\s*:([A-Za-z][A-Za-z_]*)(?:\s+([A-Za-z_]+))?(?:\s+:(0|[1-9][0-9]*))?', current_line_text)
-		if not m: raise grammar.DefinitionError('Unable to analyze overall pattern/action/parameter/(rank) structure at line %d.'%line_number)
-		pattern, action, parameter, rank_string = m.groups()
-		rank = int(rank_string) if rank_string else 0
-		note_pattern(pattern)
-		for trail, list_of_patterns in pending_patterns.items():
-			rule_id = foundation.allocate(scan_actions, (action, parameter, trail, line_number))
-			for bol, expression in list_of_patterns:
-				src = nfa.new_node(rank)
-				dst = nfa.new_node(rank)
-				for q,b in zip(nfa.condition(current_pattern_group), bol):
-					if b: nfa.link_epsilon(q, src)
-				nfa.final[dst] = rule_id
-				regular.Encoder(nfa, rank, env).visit(expression, src, dst)
-		pending_patterns.clear()
+			pattern_text = current_line_text[:-1].strip()
+			if re.search(r'\s', pattern_text): raise grammar.DefinitionError('Unable to analyze pattern/same-as-next structure at line %d.')
+			note_pattern(pattern_text)
+		else:
+			m = re.fullmatch(r'(.*?)\s*:([A-Za-z][A-Za-z_]*)(?:\s+([A-Za-z_]+))?(?:\s+:(0|[1-9][0-9]*))?', current_line_text)
+			if not m: raise grammar.DefinitionError('Unable to analyze overall pattern/action/parameter/(rank) structure at line %d.'%line_number)
+			pattern_text, action, parameter, rank_string = m.groups()
+			rank = int(rank_string) if rank_string else 0
+			note_pattern(pattern_text)
+			for trail, list_of_patterns in pending_patterns.items():
+				rule_id = foundation.allocate(scan_actions, ScanRule(line_number, trail, action, parameter))
+				for bol, expression in list_of_patterns:
+					src = nfa.new_node(rank)
+					dst = nfa.new_node(rank)
+					for q,b in zip(nfa.condition(current_pattern_group), bol):
+						if b: nfa.link_epsilon(q, src)
+					nfa.final[dst] = rule_id
+					expression.tour(regular.Encoder(nfa, rank, env), src, dst)
+			pending_patterns.clear()
 		pass
 	
 	def precedence(): ebnf.read_precedence_line(current_line_text, line_number)
@@ -177,7 +191,7 @@ def compile_string(document:str, *, method) -> IntermediateForm:
 			nonlocal current_pattern_group
 			current_pattern_group = tokens[1] if len(tokens)>1 else interfaces.DEFAULT_INITIAL_CONDITION
 			return patterns
-		if head == 'precedence': return precedence
+		if head in ('precedence', 'declarations'): return precedence
 		if head == 'productions':
 			ebnf.current_head = None
 			for t in tokens[1:]:
@@ -216,7 +230,7 @@ def compile_string(document:str, *, method) -> IntermediateForm:
 	# Here begins the outermost layer of grammar definition parsing, which is to comprehend the
 	# structure of a supplied mark-down document just enough to extract headers and code-blocks.
 	section, in_code, line_number = None, False, 0
-	for current_line_text in document.splitlines(keepends=False):
+	for current_line_text in document.content.splitlines(keepends=False):
 		line_number += 1
 		if in_code:
 			current_line_text = current_line_text.strip()
